/*
	This file is part of ldraw.

	ldraw is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ldraw is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with ldraw.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2008 Lucas Hermann Negri
*/

#include "ldraw.h"

/* Real constructors */
static void priv_context_new(lua_State* L, cairo_t* cr, BOOL need_unref);
static void priv_surface_new(lua_State* L, cairo_surface_t* sf);
static void priv_pattern_new(lua_State* L, cairo_pattern_t* pt);
static void priv_matrix_new(lua_State* L, cairo_matrix_t* mt);

/**
 * Creates a double array with the contens of a table.
 * The caller is the owner of the array.
 * 
 * @param L lua_State to use
 * @param tpos Position of the table in the stack
 * @param size Size of the table
 * @return Newly allocated double array
 */ 
static double* tableToDoubleArray(lua_State* L, int tpos, int size)
{
	int i;
	double* array = malloc(sizeof(double) * size);
	
	for(i = 1; i <= size; ++i)
	{
		lua_rawgeti(L, tpos, i);
		array[i - 1] = lua_tonumber(L, -1);
	}
	
	return array;	
}

int ldraw_version(lua_State* L)
{	
	lua_pushliteral(L, LIB_VERSION);
	lua_pushstring(L, cairo_version_string());
	lua_pushstring(L, CAIRO_VERSION_STRING);
	
	return 3;
}

void priv_context_new(lua_State* L, cairo_t* cr, BOOL need_unref)
{
	#ifdef IDEBUG
	fprintf(stderr, "Creating a cairo_t!\n");
	#endif

	size_t tam = sizeof(Context);
	Context* context = lua_newuserdata(L, tam);
	context->pointer = cr;
	context->need_unref = need_unref;

	/* Set the metatable */
	luaL_getmetatable(L, "__MtCoNtExT");
	lua_setmetatable(L, -2);
}

void priv_surface_new(lua_State* L, cairo_surface_t* sf)
{
	#ifdef IDEBUG
	fprintf(stderr, "Creating a cairo_surface_t!\n");
	#endif

	size_t tam = sizeof(Context);
	Context* context = lua_newuserdata(L, tam);
	context->pointer = sf;
	context->need_unref = TRUE;

	/* Set the metatable */
	luaL_getmetatable(L, "__MtSuRfAcE");
	lua_setmetatable(L, -2);
}

void priv_matrix_new(lua_State* L, cairo_matrix_t* mt)
{
	#ifdef IDEBUG
	fprintf(stderr, "Creating a cairo_matrix_t!\n");
	#endif

	size_t tam = sizeof(Context);
	Context* context = lua_newuserdata(L, tam);
	context->pointer = (void*)mt;
	context->need_unref = TRUE;

	/* Set the metatable */
	luaL_getmetatable(L, "__MtMaTrIx");
	lua_setmetatable(L, -2);
}

void priv_pattern_new(lua_State* L, cairo_pattern_t* sf)
{
	#ifdef IDEBUG
	fprintf(stderr, "Creating a cairo_pattern_t!\n");
	#endif

	size_t tam = sizeof(Context);
	Context* context = lua_newuserdata(L, tam);
	context->pointer = (void*)sf;
	context->need_unref = TRUE;

	/* Set the metatable */
	luaL_getmetatable(L, "__MtPaTtErN");
	lua_setmetatable(L, -2);
}

/* Begin of cairo functions */

int ldraw_context_create(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* object	= lua_touserdata(L, 1);

	cairo_t* sf = cairo_create((cairo_surface_t*)object->pointer);
	priv_context_new(L, sf, TRUE);

	return 1;
}

/*
 * Optional integration with lgui
 */
#ifdef HAVE_GTK
int ldraw_context_create_from_drawable(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* object	= lua_touserdata(L, 1);

	cairo_t* cr = gdk_cairo_create(GDK_DRAWABLE(object->pointer));
	priv_context_new(L, cr, TRUE);

	return 1;
}

int ldraw_context_get_from_print_context(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Object* object	= lua_touserdata(L, 1);

	cairo_t* cr = gtk_print_context_get_cairo_context(GTK_PRINT_CONTEXT(object->pointer));
	priv_context_new(L, cr, FALSE);

	return 1;
}

int ldraw_pango_context_layout_path(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	pango_cairo_layout_path((cairo_t*)obj1->pointer, PANGO_LAYOUT(obj2->pointer));

	return 0;
}

int ldraw_pango_context_show_layout(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	pango_cairo_show_layout((cairo_t*)obj1->pointer, PANGO_LAYOUT(obj2->pointer));
	return 0;
}

int ldraw_pango_update_layout(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	pango_cairo_update_layout((cairo_t*)obj1->pointer, PANGO_LAYOUT(obj2->pointer));

	return 0;
}

int ldraw_context_set_source_color(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	gdk_cairo_set_source_color((cairo_t*)obj1->pointer, (GdkColor*)obj2->pointer);

	return 0;
}

int ldraw_context_set_source_pixbuf(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	Object* obj1 = lua_touserdata(L, 1);
	Object* obj2 = lua_touserdata(L, 2);

	gdk_cairo_set_source_pixbuf((cairo_t*)obj1->pointer, GDK_PIXBUF(obj2->pointer),
		lua_tointeger(L, 3), lua_tointeger(L, 4));

	return 0;
}

#endif

int ldraw_context_status(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	
	Context* context = lua_touserdata(L, 1);
	int res = cairo_status((cairo_t*)context->pointer);
	lua_pushnumber(L, res);
	
	return 1;
}

int ldraw_context_set_dash(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TTABLE);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	
	int size = lua_tointeger(L, 3);
	Context* obj = lua_touserdata(L, 1);
	double* array = tableToDoubleArray(L, 2, size);
	cairo_set_dash((cairo_t*)obj->pointer, array, size, lua_tonumber(L, 4));
	free(array);
	
	return 0;
}

int ldraw_context_set_source_surface(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	Context* surface =  lua_touserdata(L, 2);

	cairo_set_source_surface((cairo_t*)context->pointer, (cairo_surface_t*)surface->pointer,
	lua_tonumber(L, 3), lua_tonumber(L, 4));

	return 0;
}

int ldraw_context_set_line_cap(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_line_cap((cairo_t*)context->pointer, (cairo_line_cap_t)lua_tointeger(L, 2));

	return 0;
}

int ldraw_context_set_line_join(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_line_cap((cairo_t*)context->pointer, (cairo_line_join_t)lua_tointeger(L, 2));

	return 0;
}

int ldraw_context_set_source_rgb(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_source_rgb((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4));

	return 0;
}

int ldraw_context_set_source_rgba(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_source_rgba((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5));

	return 0;
}
int ldraw_context_clip(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	cairo_clip((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_save(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	cairo_save((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_restore(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	cairo_restore((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_paint(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	cairo_paint((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_paint_with_alpha(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_paint_with_alpha((cairo_t*)context->pointer, lua_tonumber(L, 2));

	return 0;
}

int ldraw_context_select_font_face(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TSTRING);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_select_font_face((cairo_t*)context->pointer, lua_tostring(L, 2),
	(cairo_font_slant_t)lua_tointeger(L, 3),  (cairo_font_weight_t)lua_tointeger(L, 4));

	return 0;
}

int ldraw_context_set_font_size(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_font_size((cairo_t*)context->pointer, lua_tonumber(L, 2));

	return 0;
}

int ldraw_context_move_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_move_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_close_path(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);
	
	cairo_close_path((cairo_t*)context->pointer);
	
	return 0;
}

int ldraw_context_line_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_line_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_curve_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);
	luaL_checktype(L, 7, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_curve_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6), lua_tonumber(L, 7));

	return 0;
}

int ldraw_context_rel_move_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_rel_move_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_rel_line_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_rel_line_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_rel_curve_to(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);
	luaL_checktype(L, 7, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_rel_curve_to((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6), lua_tonumber(L, 7));

	return 0;
}

int ldraw_context_arc(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_arc((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6));

	return 0;
}

int ldraw_context_arc_negative(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_arc_negative((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6));

	return 0;
}

int ldraw_context_show_text(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TSTRING);

	Context* context =  lua_touserdata(L, 1);
	cairo_show_text((cairo_t*)context->pointer, lua_tostring(L, 2));

	return 0;
}

int ldraw_context_destroy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	if(context->need_unref)
	{
		context->need_unref = FALSE;
		cairo_destroy((cairo_t*)context->pointer);
	}
	
	#ifdef IDEBUG
	fprintf(stderr, "Context destroyed!\n");
	#endif

	return 0;
}

int ldraw_context_set_line_width(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_line_width((cairo_t*)context->pointer, lua_tonumber(L, 2));

	return 0;
}

int ldraw_context_rectangle(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_rectangle((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3),
	lua_tonumber(L, 4), lua_tonumber(L, 5));

	return 0;
}

int ldraw_context_stroke(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	cairo_stroke((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_fill(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	cairo_fill((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_set_antialias(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_set_antialias((cairo_t*)context->pointer, (cairo_antialias_t)lua_tointeger(L, 2));

	return 0;
}

int ldraw_context_get_antialias(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	int res = (int)cairo_get_antialias((cairo_t*)context->pointer);
	lua_pushnumber(L, res);

	return 1;
}

int ldraw_context_translate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_translate((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_scale(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_scale((cairo_t*)context->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	return 0;
}

int ldraw_context_rotate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_rotate((cairo_t*)context->pointer, lua_tonumber(L, 2));

	return 0;
}

int ldraw_surface_status(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context = lua_touserdata(L, 1);
	int res = cairo_surface_status((cairo_surface_t*)context->pointer);
	lua_pushnumber(L, res);

	return 1;
}

int ldraw_surface_destroy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	if(context->need_unref)
	{
		context->need_unref = FALSE;
		cairo_surface_destroy((cairo_surface_t*)context->pointer);
	}
	
	return 0;
}

int ldraw_image_surface_create_from_png(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	cairo_surface_t* sf = cairo_image_surface_create_from_png(lua_tostring(L, 1));
	priv_surface_new(L, sf);

	return 1;
}

int ldraw_surface_create_pdf(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	
	cairo_surface_t* sf = cairo_pdf_surface_create(lua_tostring(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3));
	priv_surface_new(L, sf);

	return 1;
}

int ldraw_surface_create_svg(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	
	cairo_surface_t* sf = cairo_svg_surface_create(lua_tostring(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3));
	priv_surface_new(L, sf);

	return 1;
}

int ldraw_surface_create_ps(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TSTRING);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	
	cairo_surface_t* sf = cairo_ps_surface_create(lua_tostring(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3));
	priv_surface_new(L, sf);

	return 1;
}

int ldraw_surface_write_to_png(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TSTRING);
	Context* context =  lua_touserdata(L, 1);

	cairo_surface_write_to_png((cairo_surface_t*)context->pointer, lua_tostring(L, 2));

	return 0;
}

int ldraw_image_surface_create(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	cairo_surface_t* sf = cairo_image_surface_create((cairo_format_t)lua_tointeger(L, 1),
	lua_tointeger(L, 2), lua_tointeger(L, 3));
	priv_surface_new(L, sf);

	return 1;
}

int ldraw_pattern_destroy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	if(context->need_unref)
	{
		context->need_unref = FALSE;
		cairo_pattern_destroy((cairo_pattern_t*)context->pointer);
	}

	return 0;
}

int ldraw_pattern_create_linear(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	cairo_pattern_t* pt = cairo_pattern_create_linear(lua_tonumber(L, 1), lua_tonumber(L, 2),
	lua_tonumber(L, 3), lua_tonumber(L, 4));

	priv_pattern_new(L, pt);

	return 1;
}

int ldraw_pattern_create_radial(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TNUMBER);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);

	cairo_pattern_t* pt = cairo_pattern_create_radial(lua_tonumber(L, 1), lua_tonumber(L, 2),
	lua_tonumber(L, 3), lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6));

	priv_pattern_new(L, pt);

	return 1;
}

int ldraw_pattern_add_color_stop_rgb(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);

	cairo_pattern_add_color_stop_rgb((cairo_pattern_t*)context->pointer, lua_tonumber(L, 2),
	lua_tonumber(L, 3), lua_tonumber(L, 4), lua_tonumber(L, 5));

	return 0;
}

int ldraw_pattern_add_color_stop_rgba(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	cairo_pattern_add_color_stop_rgba((cairo_pattern_t*)context->pointer, lua_tonumber(L, 2),
	lua_tonumber(L, 3), lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6));

	return 0;
}

int ldraw_context_set_source(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	Context* pattern =  lua_touserdata(L, 2);

	cairo_set_source((cairo_t*)context->pointer, (cairo_pattern_t*)pattern->pointer);

	return 0;
}

int ldraw_context_mask(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	Context* pattern =  lua_touserdata(L, 2);

	cairo_mask((cairo_t*)context->pointer, (cairo_pattern_t*)pattern->pointer);

	return 0;
}

int ldraw_context_mask_surface(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);
	Context* surface =  lua_touserdata(L, 2);

	cairo_mask_surface((cairo_t*)context->pointer, (cairo_surface_t*)surface->pointer,
	lua_tonumber(L, 3), lua_tonumber(L, 4));

	return 0;
}

int ldraw_surface_get_size(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	lua_pushnumber(L, cairo_image_surface_get_width((cairo_surface_t*)context->pointer));
	lua_pushnumber(L, cairo_image_surface_get_height((cairo_surface_t*)context->pointer));

	return 2;
}

int ldraw_context_stroke_preserve(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	cairo_stroke_preserve((cairo_t*)context->pointer);

	return 0;
}

int ldraw_context_fill_preserve(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	cairo_fill_preserve((cairo_t*)context->pointer);

	return 0;
}

int ldraw_matrix_create(lua_State* L)
{
	cairo_matrix_t* mt = malloc(sizeof(cairo_matrix_t));
	priv_matrix_new(L, mt);

	return 1;
}

int ldraw_matrix_destroy(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* context =  lua_touserdata(L, 1);

	if(context->need_unref)
	{
		context->need_unref = FALSE;
		free(context->pointer);
	}
	
	#ifdef IDEBUG
	fprintf(stderr, "Matrix destroyed!\n");
	#endif

	return 0;
}

int ldraw_matrix_init(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);
	luaL_checktype(L, 4, LUA_TNUMBER);
	luaL_checktype(L, 5, LUA_TNUMBER);
	luaL_checktype(L, 6, LUA_TNUMBER);
	luaL_checktype(L, 7, LUA_TNUMBER);

	Context* context =  lua_touserdata(L, 1);

	cairo_matrix_init((cairo_matrix_t*)context->pointer, lua_tonumber(L, 2),
	lua_tonumber(L, 3), lua_tonumber(L, 4), lua_tonumber(L, 5), lua_tonumber(L, 6),
	lua_tonumber(L, 7));

	#ifdef IDEBUG
	cairo_matrix_t* mt = context->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_context_transform(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Context* context =  lua_touserdata(L, 1);
	Context* matrix =  lua_touserdata(L, 2);

	cairo_transform((cairo_t*)context->pointer, (cairo_matrix_t*)matrix->pointer);

	return 0;
}

int ldraw_matrix_init_identity(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* matrix =  lua_touserdata(L, 1);
	cairo_matrix_init_identity((cairo_matrix_t*)matrix->pointer);

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_translate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* matrix =  lua_touserdata(L, 1);
	cairo_matrix_translate((cairo_matrix_t*)matrix->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_rotate(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);

	Context* matrix =  lua_touserdata(L, 1);
	cairo_matrix_rotate((cairo_matrix_t*)matrix->pointer, lua_tonumber(L, 2));

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_scale(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* matrix =  lua_touserdata(L, 1);
	cairo_matrix_scale((cairo_matrix_t*)matrix->pointer, lua_tonumber(L, 2), lua_tonumber(L, 3));

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_invert(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);

	Context* matrix =  lua_touserdata(L, 1);
	cairo_matrix_invert((cairo_matrix_t*)matrix->pointer);

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_multiply(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);

	Context* matrix1 =  lua_touserdata(L, 1);
	Context* matrix2 =  lua_touserdata(L, 2);
	Context* matrix3 =  lua_touserdata(L, 3);

	cairo_matrix_multiply((cairo_matrix_t*)matrix1->pointer, (cairo_matrix_t*)matrix2->pointer,
	(cairo_matrix_t*)matrix3->pointer);

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix1->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	return 0;
}

int ldraw_matrix_transform_distance(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* matrix =  lua_touserdata(L, 1);
	double v1 = lua_tonumber(L, 2);
	double v2 = lua_tonumber(L, 3);

	cairo_matrix_transform_distance((cairo_matrix_t*)matrix->pointer, &v1, &v2);

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	lua_pushnumber(L, v1);
	lua_pushnumber(L, v2);

	return 2;
}

int ldraw_matrix_transform_point(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TNUMBER);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Context* matrix =  lua_touserdata(L, 1);
	double v1 = lua_tonumber(L, 2);
	double v2 = lua_tonumber(L, 3);

	cairo_matrix_transform_point((cairo_matrix_t*)matrix->pointer, &v1, &v2);

	#ifdef IDEBUG
	cairo_matrix_t* mt = matrix->pointer;
	fprintf(stderr, "Matrix values: xx: %f, xy: %f, x0: %f, yy: %f, yx: %f, y0: %f!\n", mt->xx, mt->xy, mt->x0,
	mt->yy, mt->yx, mt->y0);
	#endif

	lua_pushnumber(L, v1);
	lua_pushnumber(L, v2);

	return 2;
}

int ldraw_context_tostring(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* obj = lua_touserdata(L, 1);
	
	char name[31];
	snprintf(name, 30, "cairo_t: %p", obj->pointer);
	lua_pushstring(L, name);
	
	return 1;
}

int ldraw_surface_tostring(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* obj = lua_touserdata(L, 1);
	
	char name[31];
	snprintf(name, 30, "cairo_surface_t: %p", obj->pointer);
	lua_pushstring(L, name);
	
	return 1;
}

int ldraw_matrix_tostring(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* obj = lua_touserdata(L, 1);
	
	char name[31];
	snprintf(name, 30, "cairo_matrix_t: %p", obj->pointer);
	lua_pushstring(L, name);
	
	return 1;
}

int ldraw_pattern_tostring(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	Context* obj = lua_touserdata(L, 1);
	
	char name[31];
	snprintf(name, 30, "cairo_pattern_t: %p", obj->pointer);
	lua_pushstring(L, name);
	
	return 1;
}
